from netgen.csg import *
from netgen.occ import *
from ngsolve import *
from ngsolve.webgui import *
import numpy as np
import netgen.gui

# Define simple geometry
geo = CSGeometry()
cube = OrthoBrick( Pnt(0,0,0), Pnt(900,900,900) ) # brick with 900mm x 900mm x 900mm
cube.bc("first")
geo.Add (cube) #,maxh=0.2) 

mesh_netgen = geo.GenerateMesh(maxh=300) # global mesh size

# name of material (1 based index)
mesh_netgen.SetMaterial(1, "product")

mesh = Mesh(mesh_netgen)

print("materials  = ", mesh.GetMaterials())
print("boundaries = ", mesh.GetBoundaries())

print(mesh.Elements())

# Finite element space
h1 = H1(mesh, order=1)
u,v = h1.TnT()

print(h1.ndof) # number of unknowns in this space

# Set initial temperature
T_0 = 260 # initial_value
gfu = GridFunction(h1)  # solution
gfu.Set(T_0) #set intial condition
gfu_old = gfu

if os.path.isfile("output.html"): os.remove("output.html")
if os.path.isfile("init_sol.html"): os.remove("init_sol.html")
Draw(gfu)
os.rename('output.html', "init_sol.html")

# Define material parameters
coeffs_lam_0 = 146.9 # W/m/K
coeffs_lam_1 = -0.05 # W/m/K^2

coeffs_rho_c_0 = 2.3*1e6 # kg/m^3*J/kg/K = J/m^3/K
coeffs_rho_c_1 = 0.0011*1e6 # J/m^3/K^2

# factors to simulate an inhomogeneous heat conductivity
scal_x = 1
scal_y = 0.01 #1
scal_z = 0.1 #1

coeff_a = CoefficientFunction(((coeffs_lam_0+coeffs_lam_1*gfu_old)*scal_x/1000,0,0,0,(coeffs_lam_0+coeffs_lam_1*gfu_old)*scal_y/1000,0,0,0,(coeffs_lam_0+coeffs_lam_1*gfu_old)*scal_z/1000),dims=(3,3)) # heat conductivity in W/mm/K
coeff_m = CoefficientFunction((coeffs_rho_c_0+coeffs_rho_c_1*gfu_old)/1000000000) # density in kg/mm^3 times specific heat capacity in J/kg/K
    
a = BilinearForm((coeff_a * grad(u)) * grad(v) * dx) # BilinearForm
m = BilinearForm((coeff_m * u) * v * dx)

# create constant input over whole surface (neumann condition)
l2surf = SurfaceL2(mesh, order=0)
surffunc = GridFunction(l2surf)
surf_data = np.full([mesh.GetNE(BND)],20000/1000000) # constant heat flow into the brick, W/mm^2

surffunc.vec.FV().NumPy()[:] = surf_data

if os.path.isfile("output.html"): os.remove("output.html")
if os.path.isfile("indata.html"): os.remove("indata.html")
Draw(surffunc, mesh, "indata", draw_vol=False)
os.rename('output.html', "indata.html")

# use as neumann boundary condition
f = LinearForm(surffunc * v * ds)
f.Assemble()


# time integration
dt = 10
t_end = 20000 #[s]

res = gfu.vec.CreateVector()

t = 0.0

min_temp = np.full([round(t_end/dt)],0) # vector for minimum temperatures

ind = 0
while t < t_end:  
    a.Assemble() # old solution gfu_old is used to evaluate nonlinear material parameters
    m.Assemble()
    
    mstar = m.mat.CreateMatrix()
    mstar.AsVector().data = m.mat.AsVector() + dt * a.mat.AsVector()
    invmstar = mstar.Inverse(freedofs=h1.FreeDofs(), inverse="sparsecholesky")
    

    res.data = f.vec - a.mat*gfu.vec
    
    gfu_old.vec.data = gfu.vec
    gfu.vec.data += dt*invmstar * res  
    
    t=round(t+dt,6) #round due to machine precision epsilon
    print('Time: ' + str(t))
    Redraw(blocking=True)
    
    min_temp[ind] = min(gfu.vec)
    ind = ind+1
    
    if (t==10*dt or t==100*dt or t==500*dt or t==1000*dt or t==2000*dt):  # textbook example for floating point arithmetics
        # print some solutions
        stri = 'sol_t_' + str(t) +'s.html'
        if os.path.isfile("output.html"): os.remove("output.html")
        if os.path.isfile(stri): os.remove(stri)
        Draw(gfu)
        os.rename('output.html', stri)

print('minimum temperature inside brick over time:')
print(min_temp)

print('absolute minimum temperature during simulation:')
print(min(min_temp))
